require 'Nully/version'
require 'delegate'
require 'set'

# Top abstraction.
class Object
  def nully(*methods)
    options = methods[1]
    method = methods[0]

    file, line = file_and_line_from_caller(caller)
    line = line.to_i

    build_class_var(method, options)
    build_alias(method)
    module_eval(rewritten_method(method, options), file, line)
  end

  private

  def rewritten_method(method, options)
    [
      "def #{method}(#{method_def(method)})",
      "initial_result = self.__nully_#{method}(#{method_def(method)})",
      "#{options[:delegate_on_false] ? false_check(method, options) : normal_check(method, options)}",
      'end'
    ].join(';')
  end

  def normal_check(method, options)
    "return initial_result unless initial_result == nil;
     #{delegation_method(method, options)}"
  end

  def false_check(method, options)
    "return initial_result || #{delegation_method(method, options)}"
  end

  def delegation_method(method, options)
    if options[:to].is_a?(Symbol)
      if options[:skip_delegation_call]
        "self.#{options[:to]}"
      else
        "self.#{options[:to]}(#{method_def(method)})"
      end
    elsif options[:to].is_a?(Proc)
      if options[:skip_delegation_call]
        "#{class_var_name(method)}"
      else
        "#{class_var_name(method)}.call.#{method}(#{method_def(method)})"
      end
    else
      if options[:skip_delegation_call]
        "#{class_var_name(method)}"
      else
        "#{class_var_name(method)}.#{method}(#{method_def(method)})"
      end
    end
  end

  def method_def(method)
    (method =~ /[^\]]=$/) ? 'arg' : '*args, &block'
  end

  def build_class_var(method, options)
    class_variable_set(class_var_name(method), options[:to])
  end

  def build_alias(method)
    alias_method("__nully_#{method}".to_sym, method)
  end

  def class_var_name(method)
    "@@__#{make_name_safe(method.to_s)}__"
  end

  def make_name_safe(name)
    name.gsub('!', '').gsub('?', '')
  end

  def file_and_line_from_caller(caller)
    caller.first.split(':', 2)
  end
end
